/*
 * Copyright (c) 2006 Nicolas Bernard <http://www.lafraze.net/nbernard/>
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/dh.h>
#include <openssl/err.h>

#include "dtlsplex.h"
#include "netmisc.h"
#include "peer.h"




static SSL*
SSL_create(SSL_CTX* ctx, BIO* conn, bool client)
{
	assert(ctx != NULL);
	assert(conn != NULL);

	SSL *ssl = NULL;
	ssl = SSL_new(ctx);
	if (ssl == NULL) {
		return NULL;
	}
	if (client)
		SSL_set_connect_state(ssl);
	else
		SSL_set_accept_state(ssl);

	SSL_set_bio(ssl, conn, conn);

	return ssl;
}

static SSL_CTX*
createctx(void)
{
	SSL_CTX* ctx = SSL_CTX_new(DTLSv1_method());
	if (ctx == NULL) {
		ERR_print_errors_fp(stderr);
		return NULL;
	}

	int err = SSL_CTX_set_cipher_list(ctx, "HIGH:MEDIUM:aNULL");
	if (err != 1) {
		fprintf(stderr, "DTLS: unable to load ciphers, exiting\n");
		exit(1);
	}

	FILE* paramfile = fopen("dh_param_1024.pem", "r");
	if (paramfile == NULL) {
		fprintf(stderr, "Error opening the DH file: %s\n",
			strerror(errno));
		return NULL;
	}

	DH* dh_1024 = PEM_read_DHparams(paramfile, NULL, NULL, NULL);
	fclose(paramfile);
	if (dh_1024 == NULL) {
		fprintf(stderr, "Error reading the DH file\n");
		return NULL;
	}

	err = SSL_CTX_set_tmp_dh(ctx, dh_1024);
	if (err != 1) {
		fprintf(stderr, "createctx: unable to set DH parameters\n");
		ERR_print_errors_fp(stderr);
		SSL_CTX_free(ctx);
		ctx = NULL;
		return NULL;
	}
	return ctx;
}

static struct dtls_peer*
dtlsnewpeer(const struct sockaddr *so, bool client)
{
	assert(so != NULL);

	static SSL_CTX *ctx = NULL;
	int err = 0;
	if (ctx == NULL) {
		ctx = createctx();
		if (ctx == NULL) {
			fprintf(stderr, "dtlsnewpeer: unable to create CTX\n");
			return NULL;
		}
	}

	struct dtls_peer p;

	err = BIO_new_bio_pair(&p.bio, MTU, &p._b2, MTU);
	if (err != 1) {
		fprintf(stderr, "dtlsnewpeer: unable to create bio pair\n");
		ERR_print_errors_fp(stderr);
		return NULL;
	}

	p.ssl = SSL_create(ctx, p._b2, client);
	if (p.ssl == NULL) {
		BIO_free(p.bio);
		BIO_free(p._b2);
		return NULL;
	}

	return addpeer(&p, so);
}

ssize_t
dtlssendto(int s, const void *msg, size_t len, int flags,
	   const struct sockaddr *to, socklen_t tolen)
{
	ssize_t ret = -1;
	int err = 0;
	unsigned char* tbuf[MTU] = {0};
	ssize_t retv = -10;

	if (len > MTU) {
		/* what about TLS encapsulation size? */
		errno = EMSGSIZE;
		return -2;
	}

	struct dtls_peer *p = getpeer(to, tolen);
	if (p == NULL) {
		fprintf(stderr, "unknown peer!\n");
		p = dtlsnewpeer(to, true);
		if (p == NULL) {
			fprintf(stderr, "dtlssendto: peer creation failed\n");
			return -3;
		}
		err = SSL_connect(p->ssl);
		retv = 0;
	} else {
		err = SSL_write(p->ssl, msg, len);
		if (err != len) {
			int ret = SSL_get_error(p->ssl, err);
			fprintf(stderr, "dtlssendto: SSL_write/connect returned %d"
				" (len was %d): error %d\n", err, len, ret);
			ERR_print_errors_fp(stderr);
		}
		retv = len;
	}

	ret = BIO_read(p->bio, tbuf, MTU);
	if (ret <= 0) {
		if (ret < 0)
			fprintf(stderr, "dtlssendto: BIO_read error (%d)\n",
				ret);
		return 0;
	}

	err = sendto(s, tbuf, ret, flags, to, tolen);
	if (err != ret) {
		fprintf(stderr, "dtlssendto: sendto: %s\n",
			strerror(errno));
		return -1;
	}
	return retv;
}

static struct dtls_peer*
socktobssl(int s, int flags)
{
	ssize_t ret = -1;
	unsigned char* tbuf[MTU] = {0};
	struct sockaddr_storage ifrom;
	socklen_t ifromlen = sizeof ifrom;
	ret = recvfrom(s, tbuf, MTU, flags, (struct sockaddr*) &ifrom,
		       &ifromlen);
	if (ret == -1) {
		fprintf(stderr, "dtlsrecvfrom: recvfrom: %s\n",
			strerror(errno));
		return NULL;
	}

	struct dtls_peer *p = getpeer((struct sockaddr*) &ifrom, ifromlen);
	if (p == NULL) {
		fprintf(stderr, "unknown peer!\n");
		p = dtlsnewpeer((struct sockaddr*) &ifrom, false);
		if (p == NULL)
			return NULL;
	}

#ifdef DBG_DTLS
	fprintf(stderr, "Writing %d octet(s) to the bio\n", ret);
#endif
	int nret = BIO_write(p->bio, tbuf, ret);
	assert(nret == ret);
	int needed = BIO_get_read_request(p->bio);
	if (needed > 0) {
		fprintf(stderr, "dtlsrecvfrom: Still %d octet(s) needed... \n",
			needed);
		return NULL;
	}
	return p;
}

static int
ssltobuf(int s, void* buf, size_t len, int flags,
	 struct sockaddr *from, socklen_t *fromlen, struct dtls_peer* p)
{
	assert(p != NULL);

	int ret2 = SSL_read(p->ssl, buf, len);
	if (ret2 < 0) {
		int err = SSL_get_error(p->ssl, err);
		fprintf(stderr, "SSL_read: returned %d, error %d\n", ret2, err);
		ERR_print_errors_fp(stderr);
	}

	int retb = 0;
	if ((retb = BIO_ctrl_pending(p->bio)) > 0) {
#ifdef DBG_DTLS
		fprintf(stderr, "sending pending data (%d octets to send)\n",
			retb);
#endif
		unsigned char* tbuf[MTU] = {0};
		int ret3 = BIO_read(p->bio, tbuf, MTU);
		if (ret3 != retb)
			fprintf(stderr, "dtlsrecvfrom: warning "
				"datatosend > MTU, %d != %d\n", ret3, retb);
		retb = sendto(s, tbuf, ret3, flags, (struct sockaddr*) &p->addr,
			      addrsize((struct sockaddr*) &p->addr));
		if (retb < 0) {
			fprintf(stderr, "sendto: %s\n", strerror(errno));
		}
		assert(ret3 == retb);
		assert(ret2 <= 0);
		return dtlsrecvfrom(s, buf, len, flags, from, fromlen);
	}
	assert(retb == 0);
	if (ret2 < 0)
		return -1;

	if (from != NULL && fromlen != NULL) {
		memcpy(from, &p->addr, sizeof(struct sockaddr_storage));
		*fromlen = addrsize(from);
	}
	return (ssize_t) ret2;
}

ssize_t
dtlsrecvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from,
	     socklen_t *fromlen)
{
	struct dtls_peer *p = NULL;
	p = socktobssl(s, flags);
	if (p != NULL)
		return ssltobuf(s, buf, len, flags, from, fromlen, p);
	return -1;
}
