/*
 * Misc. network related functions
 *
 * Copyright (c) 2004-2006 Nicolas Bernard
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "netmisc.h"

static int
isloopbackip4(struct sockaddr* addr)
{ /* all the 127.0.0.* addresses can be loopback */
	unsigned long int ip = 0;
	if (((struct sockaddr_in*) addr)->sin_family != AF_INET) return -1;

	ip = (((struct sockaddr_in*) addr)->sin_addr.s_addr);
#ifndef NDEBUG
	fprintf(stderr, "ip4: %ld.%ld.%ld.%ld\n", ip % 256, (ip >> 8) % 256, (ip >> 16) % 256, (ip >> 24) % 256);
#endif
	if (ip % 256 == 127 && (ip >> 8) % 256 == 0 && (ip >> 16) % 256 == 0)
		return 1;

	return 0;
}

static int
isloopbackip6(struct sockaddr* addr)
{ /* 0000:0000:0000:0000:0000:0000:0000:0001 only ? */
	/** \test test this function */
	if (((struct sockaddr_in6*) addr)->sin6_family != AF_INET6) return -1;
#ifdef __KAME__
#ifndef NDEBUG
	for (int i = 0; i < 16; i++) {
		fprintf(stderr, "%.2X", ((struct sockaddr_in6*) addr)->sin6_addr.__u6_addr.__u6_addr8[i]);
		if (i % 2)
			fprintf(stderr, ":");
	}
	fprintf(stderr, "\n");
#endif
	for (int i = 0; i < 3; i++) {
		if (((struct sockaddr_in6*) addr)->sin6_addr.__u6_addr.__u6_addr32[i])
			return 0;
	}
	if (((struct sockaddr_in6*) addr)->sin6_addr.__u6_addr.__u6_addr32[3] == htonl(1))
		return 1;
#else /* linux 2.4 ? */
	for (int i = 0; i < 3; i++) {
		if (((struct sockaddr_in6*) addr)->sin6_addr.in6_u.u6_addr32[i])
			return 0;
	}
	if (((struct sockaddr_in6*) addr)->sin6_addr.in6_u.u6_addr32[3] == htonl(1))
		return 1;
#endif

	return 0;
}

int
isloopback(struct sockaddr* addr)
{
	switch (((struct sockaddr_in*) addr)->sin_family) {
	case AF_INET:
		return isloopbackip4(addr);
	case AF_INET6:
		return isloopbackip6(addr);
	default:
		return -1;
	}
}

static int
isprivateip4(struct sockaddr* addr)
{
	if (((struct sockaddr_in*) addr)->sin_family != AF_INET) return -2;
	unsigned long int ip = 0;
	if (isloopbackip4(addr)) return 1;
	ip = (((struct sockaddr_in*) addr)->sin_addr.s_addr);

	if (ip % 256 == 10) return 2;
	if ((ip % 256 == 192) && ((ip >> 8) % 256 == 168)) return 3;
	if ((ip % 256 == 172) && ((ip >> 8) % 256 >= 16) && ((ip >> 8) % 256 <= 31)) return 4;

	return 0;
}

static int
isprivateip6(struct sockaddr* addr)
{ ///\test to test
	if (((struct sockaddr_in6*) addr)->sin6_family != AF_INET6) return -3;
	if (isloopbackip6(addr)) return 5;
#ifdef __KAME__
	if (((struct sockaddr_in6*) addr)->sin6_addr.__u6_addr.__u6_addr8[0] == 0xFE &&
	    ((struct sockaddr_in6*) addr)->sin6_addr.__u6_addr.__u6_addr8[1] & 0x80)
		return 6;
#else /* linux 2.4 ? */
	if ((((struct sockaddr_in6*) addr)->sin6_addr.in6_u.u6_addr8[0] == 0xFE) &&
	    ((struct sockaddr_in6*) addr)->sin6_addr.in6_u.u6_addr8[1] & 0x80)
		return 6;
#endif
	return 0;
}

int
isprivateip(struct sockaddr* addr)
{
	switch (((struct sockaddr_in*) addr)->sin_family) {
	case AF_INET:
		return isprivateip4(addr);
	case AF_INET6:
		return isprivateip6(addr);
	default:
		return -1;
	}
}

int
addreql(const struct sockaddr* a, const struct sockaddr* b)
{
	if (a->sa_family != b->sa_family)
			return 0;

	switch (a->sa_family) {
	case AF_INET: {
		const struct sockaddr_in *phonya, *phonyb;
#ifndef __GLIBC__
		if (a->sa_len != 16 || b->sa_len != 16)
			return -1;
#endif
		phonya = (const struct sockaddr_in *) a;
		phonyb = (const struct sockaddr_in *) b;
		if (!memcmp(&phonya->sin_addr, &phonyb->sin_addr, 4))
			return 1;
	}
		break;
	case AF_INET6: {
		const struct sockaddr_in6 *phonya, *phonyb;
#ifndef __GLIBC__
		if (a->sa_len != 28 || b->sa_len != 28)
#endif
			return -1;
		phonya = (const struct sockaddr_in6 *) a;
		phonyb = (const struct sockaddr_in6 *) b;
		if (!memcmp(&phonya->sin6_addr, &phonyb->sin6_addr, 4))
			return 1;
	}
		break;
	default:
		return -1;
	}
	return 0;
}

int
porteql(const struct sockaddr* a, const struct sockaddr* b)
{
	if (a->sa_family != b->sa_family)
			return 0;

	switch (a->sa_family) {
	case AF_INET: {
		const struct sockaddr_in *phonya, *phonyb;
#ifndef __GLIBC__
		if (a->sa_len != 16 || b->sa_len != 16)
			return -1;
#endif
		phonya = (const struct sockaddr_in *) a;
		phonyb = (const struct sockaddr_in *) b;
		if (phonya->sin_port == phonyb->sin_port)
			return 1;
	}
		break;
	case AF_INET6: {
		const struct sockaddr_in6 *phonya, *phonyb;
#ifndef __GLIBC__
		if (a->sa_len != 28 || b->sa_len != 28)
#endif
			return -1;
		phonya = (const struct sockaddr_in6 *) a;
		phonyb = (const struct sockaddr_in6 *) b;
		if (phonya->sin6_port == phonyb->sin6_port)
			return 1;
	}
		break;
	default:
		return -1;
	}
	return 0;
}

static void
echoip4(FILE* f, struct sockaddr_in *addr)
{
	fprintf(f, "%d.%d.%d.%d", addr->sin_addr.s_addr % 256,
		(addr->sin_addr.s_addr >> 8) % 256,
		(addr->sin_addr.s_addr >> 16) % 256,
		(addr->sin_addr.s_addr >> 24) % 256);
}

static void
echoip6(FILE* f, struct sockaddr_in6 *addr)
{
	for (int i = 0; i < 8; i++) {
#ifdef __KAME__
		fprintf(f, "%4X", addr->sin6_addr.__u6_addr.__u6_addr16[i]);
#else /* linux 2.4 ? */
		fprintf(f, "%4X", addr->sin6_addr.in6_u.u6_addr16[i]);
#endif
		if (i != 7)
			fprintf(f, ":");
	}
}

int
echoip(FILE* f, struct sockaddr* addr)
{
	switch (addr->sa_family) {
	case AF_INET:
		echoip4(f, (struct sockaddr_in*) addr);
		return 0;
	case AF_INET6:
		echoip6(f, (struct sockaddr_in6*) addr);
		return 0;
	default:
		return -1;
	}
}

inline unsigned int
addrsize(struct sockaddr* a)
{
#ifndef __GLIBC__
	return a->sa_len;
#else
	switch(a->sa_family) {
	case AF_INET:
		return sizeof(struct sockaddr_in);
	case AF_INET6:
		return sizeof(struct sockaddr_in6);
	default:
		return 0;
	}
#endif
}
